//----------------------------------------------------------------------------
//  EDGE Filesystem Class for Linux
//----------------------------------------------------------------------------
//
//  Copyright (c) 2003-2005  The EDGE Team.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
#include "epi.h"
#include "errors.h"
#include "strings.h"

#include "files_all.h"
#include "filesystem_all.h"

#include <dirent.h>
#ifndef NO_FNMATCH
#include <fnmatch.h>
#endif
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#define MAX_MODE_CHARS 3

#ifndef PATH_MAX
#define PATH_MAX 128
#endif

namespace epi
{

//
// epi_all_filesystem_c Constructor
//
all_filesystem_c::all_filesystem_c()
{
}

//
// all_filesystem_c Destructor
//
all_filesystem_c::~all_filesystem_c()
{
}

//
// all_filesystem_c::ConvertFlagsToMode()
//
bool all_filesystem_c::ConvertFlagsToMode(int flags, char *mode)
{
    // Must have some value in epiflags
    if (flags == 0)
        return false;

    // Check for any invalid combinations
    if ((flags & file_c::ACCESS_WRITE) && (flags & file_c::ACCESS_APPEND))
        return false;

    if (flags & file_c::ACCESS_READ)
    {
        if (flags & file_c::ACCESS_WRITE) 
            strcpy(mode, "wb+");                        // Read/Write
        else if (flags & file_c::ACCESS_APPEND)
            strcpy(mode, "ab+");                        // Read/Append
        else
            strcpy(mode, "rb");                         // Read
    }
    else
    {
        if (flags & file_c::ACCESS_WRITE)       
            strcpy(mode, "wb");                         // Write
        else if (flags & file_c::ACCESS_APPEND) 
            strcpy(mode, "ab");                         // Append
        else                                         
            return false;                              // Invalid
    }
    
    return true;
}

//
// bool all_filesystem_c::GetCurrDir()
//
bool all_filesystem_c::GetCurrDir(const char *dir, unsigned int bufsize)
{
	if (!dir)
		return false;

	return (getcwd((char *)dir, bufsize) != NULL);
}

//
// bool all_filesystem_c::SetCurrDir()
//
bool all_filesystem_c::SetCurrDir(const char *dir)
{
	if (!dir)
		return false;

	return (chdir(dir) == 0);
}

//
// bool all_filesystem_c::IsDir()
// 
bool all_filesystem_c::IsDir(const char *dir)
{
	if (!dir)
		return false;

	DIR *result = opendir(dir);

	if (result == NULL)
		return false;

	closedir(result);
	return true;
}

//
// bool all_filesystem_c::MakeDir()
//
bool all_filesystem_c::MakeDir(const char *dir)
{
	if (!dir)
		return false;

//#if defined(WIN32) ||defined(DREAMCAST)
#ifdef WIN32
	return (mkdir(dir) == 0);
#else
	return (mkdir(dir, 0775) == 0);
#endif
}

//
// bool all_filesystem_c::RemoveDir()
//
bool all_filesystem_c::RemoveDir(const char *dir)
{
	if (!dir)
		return false;

	return (rmdir(dir) == 0);
}

//
// bool all_filesystem_c::ReadDir()
//
bool all_filesystem_c::ReadDir(filesystem_dir_c *fsd, const char *dir, const char *mask)
{
	if (!dir || !fsd || !mask)
		return false;

	DIR *handle;
	char olddir[PATH_MAX];
	filesystem_direntry_s tmp_entry;
	struct dirent *fdata;

	handle = opendir(dir);
	if (handle == NULL)
		return false;

	GetCurrDir(olddir, PATH_MAX);
	SetCurrDir(dir);

	// Ensure the container is empty
	fsd->Clear();

	for (;;)
	{
		fdata = readdir(handle);
		if (fdata == NULL)
			break;

#ifndef NO_FNMATCH
		if (fnmatch(mask, fdata->d_name, 0) != 0)
			continue;
#endif

		struct stat finfo;

		if (stat(fdata->d_name, &finfo) != 0)
			continue;
		
		tmp_entry.name = new string_c(fdata->d_name);
		tmp_entry.size = finfo.st_size;
		tmp_entry.dir = S_ISDIR(finfo.st_mode) ?true:false;

		if (!fsd->AddEntry(&tmp_entry))
		{
			closedir(handle);
			delete tmp_entry.name;
			SetCurrDir(olddir);
			return false;
		}
	}

	SetCurrDir(olddir);
	closedir(handle);

	return true;
}

//
// bool all_filesystem_c::Access()
//
bool all_filesystem_c::Access(const char *name, unsigned int flags)
{
    char mode[MAX_MODE_CHARS];
    FILE *fp;

    if (!ConvertFlagsToMode(flags, mode))
        return false;

    fp = fopen(name, mode);
    if (!fp)
        return false;

    fclose(fp);
    return true;
}

//
// bool all_filesystem_c::Close()
//
bool all_filesystem_c::Close(file_c *file)
{
    FILE *fp;

    if (!file)
        return false;

    if (file->GetType() == file_c::TYPE_DISK)
    {
        all_file_c *diskfile = (all_file_c*)file;

        fp = diskfile->GetDescriptor();
        if (fp)
            fclose(fp);

		//
		// We nullify the filesystem and filehandle to 
		// NULL: the prevents recursive Close() calls
		// since the diskfile deconstructor will
		// attempt to close the file if the filesystem
		// exists.
		// 
		diskfile->Setup(NULL, NULL);
    }

    delete file;
	return true;
}

//
// bool all_filesystem_c::Copy()
//
bool all_filesystem_c::Copy(const char *dest, const char *src)
{
	file_c *dest_file;
	file_c *src_file;
	bool ok;
	
	unsigned char *buf;
	int cpsize, size;
	
	dest_file = NULL;
	src_file = NULL;
	buf = NULL;
	ok = true;

#if 0
	try
#endif
	{
		src_file = Open(src, file_c::ACCESS_READ);
		if (!src_file)
		{
#if 0
			throw error_c(EPI_ERRGEN_FOOBAR);
#else
			return false;
#endif
		}	
	
		dest_file = Open(dest, file_c::ACCESS_WRITE);
		if (!dest_file)
		{
#if 0
			throw error_c (EPI_ERRGEN_FOOBAR);
#else
			Close(src_file);
			return false;
#endif
		}
		
		buf = new unsigned char[COPY_BUF_SIZE];
		if (!buf)
		{
#if 0
        		throw error_c(EPI_ERRGEN_FOOBAR);
#else
			Close(dest_file);
			Close(src_file);
			return false;
#endif
		}

		size = src_file->GetLength();
		while (size > 0)
		{
			if (size > COPY_BUF_SIZE)
				cpsize = COPY_BUF_SIZE;
			else
				cpsize = size;
			
			if (src_file->Read(buf, cpsize) != (unsigned int)cpsize)
			{
#if 0
				throw error_c(EPI_ERRGEN_FOOBAR);
#else
				delete [] buf;
				Close(dest_file);
				Close(src_file);
				return false;
#endif
			}
	
			if (dest_file->Write(buf, cpsize) != (unsigned int)cpsize)
			{
#if 0
				throw error_c(EPI_ERRGEN_FOOBAR);
#else
				delete [] buf;
				Close(dest_file);
				Close(src_file);
				return false;
#endif
			}
	
			size -= cpsize;
		}
	}
#if 0
	catch(epi::error_c err)
	{
		ok = false;
	}
#endif

	if (src_file)
		Close(src_file);

	if (dest_file)
		Close(dest_file);

	if (buf)
		delete [] buf;

	return ok;
}

//
// bool all_filesystem_c::Delete()
//
bool all_filesystem_c::Delete(const char *name)
{
    if (!name)
        return false;

    return !(unlink(name))?true:false;
}

//
// file_c* all_filesystem_c::Open()
//
file_c* all_filesystem_c::Open(const char *name, 
                                 unsigned int flags)
{
    char mode[MAX_MODE_CHARS];
    all_file_c *file;
    FILE *fp;

    if (!name)
        return NULL;

    if (!ConvertFlagsToMode(flags, mode))
        return NULL;

    fp = fopen(name, mode);
    if (!fp)
        return NULL;

    file = new all_file_c;
    if (!file)
    {
        fclose(fp);
        return NULL;
    }

    file->Setup(this, fp);
    return file;
}

//
// bool all_filesystem_c::Rename()
//
bool all_filesystem_c::Rename(const char *oldname, 
                                        const char *newname)
{
	if (!oldname || !newname)
	{
		return false;
	}

	return (bool)(rename(oldname, newname) != -1);
}

};
