
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "mem.h"
#include "profiler.h"

#ifdef MEM_CACHE

//#define CACHE_SIZE (1024*1024)
//#define MIN_BITS 4
#define CACHE_SIZE ((1024*1024*2)-(1024*512))
#define MIN_BITS 3

#define get_index(D) ((D)>>MIN_BITS)

#define MAX_INDEX get_index(CACHE_SIZE)
#define MAX_SIZE 256

static unsigned char cache_buffer[CACHE_SIZE];
static unsigned char cache_index[MAX_INDEX];
static unsigned cache_actual=0;

#endif

#ifdef MEM_STATS

#define MEMSTATS_MAX_ALLOCS (128*1024)

static unsigned memstats_max=0;
static unsigned memstats_actual=0;
static void *memstats_allocs[MEMSTATS_MAX_ALLOCS];
static unsigned memstats_sizes[MEMSTATS_MAX_ALLOCS];
static unsigned memstats_n_allocs=0;

#ifdef WIN32
#include <windows.h>
#include <winbase.h>
static unsigned get_uordblks(void)
{
	MEMORYSTATUS MemStat;
	memset(&MemStat, 0, sizeof(MemStat));
	GlobalMemoryStatus(&MemStat);
	return MemStat.dwMemoryLoad;
}
#else
#define get_uordblks() ((mallinfo()).uordblks)
#endif

void *mem_malloc(size_t len, const char *msg) //, const char *msg)
{
	void *p=cache_malloc(len);
#ifdef MEM_CACHE
	if ((((unsigned)p) >= ((unsigned)&cache_buffer[0])) && (((unsigned)p) < ((unsigned)&cache_buffer[CACHE_SIZE])))
		return p;
#endif
	if (memstats_actual+len>memstats_max)
		memstats_max=memstats_actual+len;
#ifdef MEM_STATS_VERBOSE
	printf("MEM ALLOC: %X + %X = %X (%X max, %X dif) -> %p <- %s\n",memstats_actual,len,memstats_actual+len,memstats_max,get_uordblks() - memstats_actual,p,msg);
#endif
	memstats_actual+=len;
	if (memstats_n_allocs>MEMSTATS_MAX_ALLOCS)
	{
		printf("ERROR: MEM ALLOCS > %i <- %s\n",MEMSTATS_MAX_ALLOCS,msg);
		fflush(stdout);
		SDL_Quit();
		exit(0);
	}
	memstats_allocs[memstats_n_allocs]=p;
	memstats_sizes[memstats_n_allocs++]=len;
	return p;
}

void *mem_calloc(size_t nmemb, size_t size, const char *msg)
{
	unsigned len=size*nmemb;
	void *ret=mem_malloc(len,msg);
	if (ret)
		memset(ret,0,len);
	return ret;
}


void mem_free(void *ptr, const char *msg)
{
	int i;
#ifdef MEM_CACHE
	if ((((unsigned)ptr) >= ((unsigned)&cache_buffer[0])) && (((unsigned)ptr) < ((unsigned)&cache_buffer[CACHE_SIZE])))
	{
		cache_free(ptr);
		return;
	}
#endif
	for(i=0;i<memstats_n_allocs;i++)
		if (memstats_allocs[i]==ptr)
			break;
	if (i>=memstats_n_allocs)
	{
		printf("ERROR: FREE %p NOT FOUND <- %s\n",ptr,msg);
		fflush(stdout);
		SDL_Quit();
		exit(0);
	
	}
	cache_free(ptr);
#ifdef MEM_STATS_VERBOSE
	printf("MEM FREE: %X - %X = %X (%X max, %X dif) -> %p <- %s\n",memstats_actual,memstats_sizes[i],memstats_actual-memstats_sizes[i],memstats_max,get_uordblks() - memstats_actual,ptr,msg);
#endif
	memstats_actual-=memstats_sizes[i];
	memstats_allocs[i]=NULL;
}

void *mem_realloc(void *ptr, size_t len, const char *msg)
{
	int i;
	size_t old_len;
	void *ret;
#ifdef MEM_CACHE
	if ((((unsigned)ptr) >= ((unsigned)&cache_buffer[0])) && (((unsigned)ptr) < ((unsigned)&cache_buffer[CACHE_SIZE])))
	{
		ret=cache_realloc(ptr,len);
		if ((((unsigned)ret) >= ((unsigned)&cache_buffer[0])) && (((unsigned)ret) < ((unsigned)&cache_buffer[CACHE_SIZE])))
			return ret;
		ptr=mem_malloc(len,msg);
		memcpy(ptr,ret,len);
		free(ret);
		return ptr;
	}
#endif
	if (!ptr)
	{
		i=memstats_n_allocs++;
	}
	else
	{
		for(i=0;i<memstats_n_allocs;i++)
			if (memstats_allocs[i]==ptr)
				break;
		if (i>=memstats_n_allocs)
		{
			printf("ERROR: REALLOC %p NOT FOUND <- %s\n",ptr,msg);
			fflush(stdout);
			SDL_Quit();
			exit(0);
		}
		memstats_actual-=memstats_sizes[i];
	}
	memstats_sizes[i]=len;
	if (memstats_actual+len>memstats_max)
		memstats_max=memstats_actual+len;
	ret=cache_realloc(ptr,len);
#ifdef MEM_STATS_VERBOSE
	printf("MEM REALLOC: %X + %X = %X (%X max, %X dif) -> %p a %p <- %s\n",memstats_actual,len,memstats_actual+len,memstats_max,get_uordblks() - memstats_actual,ptr,ret,msg);
#endif
	memstats_actual+=len;
	memstats_allocs[i]=ret;
	return ret;
}

char *mem_strdup(const char *str, const char *msg)
{
	int len;
	char *result;
	if (!str)
		return NULL;
	len = (int)strlen(str);
	result = mem_malloc(len+1,msg);
	if (len>0)
		memcpy(result, str, len);
	result[len] = 0;
	return result;
}

char *mem_strndup(const char *str, size_t size, const char *msg)
{
	int len;
	char *result;
	if (!str)
		return NULL;
	len = (int)strlen(str);
	if (len>size)
		len=size;
	result = mem_malloc(len,msg);
	if (len>0)
		memcpy(result, str, len);
	result[len] = 0;
	return result;
}

void print_meminfo(const char *msg)
{
#ifdef WIN32
	MEMORYSTATUS mifo;
	memset(&mifo, 0, sizeof(mifo));
	GlobalMemoryStatus(&mifo);
#define mifo_arena mifo.dwLength
#define mifo_ordblks mifo.dwMemoryLoad
#define mifo_smblks mifo.dwTotalPhys
#define mifo_hblks mifo.dwAvailPhys
#define mifo_hblkhd mifo.dwTotalPageFile
#define mifo_usmblks mifo.dwTotalVirtual
#define mifo_fsmblks mifo.dwAvailVirtual
#define mifo_uordblks mifo.dwMemoryLoad
#define mifo_fordblks mifo.dwAvailPhys
#define mifo_keepcost mifo.dwAvailVirtual
#else
	struct mallinfo mifo=mallinfo();
#define mifo_arena mifo.arena
#define mifo_ordblks mifo.ordblks
#define mifo_smblks mifo.smblks
#define mifo_hblks mifo.hblks
#define mifo_hblkhd mifo.hblkhd
#define mifo_usmblks mifo.usmblks
#define mifo_fsmblks mifo.fsmblks
#define mifo_uordblks mifo.uordblks
#define mifo_fordblks mifo.fordblks
#define mifo_keepcost mifo.keepcost
#endif
	printf("--> MEMINFO %s (max %.8X, actual=%.8X)\n",msg,memstats_max,memstats_actual);
	printf("\tarena    = %.8X\n",mifo_arena);
	printf("\tordblks  = %.8X\n",mifo_ordblks);
	printf("\tsmblks   = %.8X\n",mifo_smblks);
	printf("\thblks    = %.8X\n",mifo_hblks);
	printf("\thblkhd   = %.8X\n",mifo_hblkhd);
	printf("\tusmblks  = %.8X\n",mifo_usmblks);
	printf("\tfsmblks  = %.8X\n",mifo_fsmblks);
	printf("\tuordblks = %.8X\n",mifo_uordblks);
	printf("\tfordblks = %.8X\n",mifo_fordblks);
	printf("\tkeepcost = %.8X\n",mifo_keepcost);
#ifdef MEM_CACHE
	int i=0,cks=0,frd=0;
	while(i<MAX_INDEX)
	{
		if (!cache_index[i])
		{
			frd++; i++;
		}
		else
		{
			cks++; i+=cache_index[i];
		}
	}
	printf("  CACHE: %i blocks, %.2f%% free\n",cks,(100.0*(double)frd)/(double)MAX_INDEX);
#endif
#undef mifo_arena
#undef mifo_ordblks
#undef mifo_smblksi
#undef mifo_hblks
#undef mifo_hblkhd
#undef mifo_usmblks
#undef mifo_fsmblks
#undef ifo_uordblks
#undef mifo_fordblks
#undef mifo_keepcost
}

#endif

#ifdef MEM_CACHE

void cache_init(void)
{
	memset((void *)&cache_buffer[0],0,CACHE_SIZE);
	memset((void *)&cache_index[0],0,MAX_INDEX);
	cache_actual=0;
}

static int get_hole(int size)
{
	int j,i;
	for(j=cache_actual;j<MAX_INDEX;j+=cache_index[j])
		if (!cache_index[j])
		{
			int max=j+size+1;
			if (max>=MAX_INDEX) break;
			for(i=j+1;i<max;i++)
				if (cache_index[i])
					break;
			if (size<(i-j))
			{
				cache_actual=(i+size+1)%MAX_INDEX;
				return j;
			}
			j=i;
		}
	if (cache_actual)
	for(j=0;j<cache_actual;j+=cache_index[j])
		if (!cache_index[j])
		{
			int max=j+size+1;
			if (max>=MAX_INDEX) break;
			for(i=j+1;i<max;i++)
				if (cache_index[i])
					break;
			if (size<(i-j))
			{
				cache_actual=(i+size+1)%MAX_INDEX;
				return j;
			}
			j=i;
		}
	return -1;
}

static __inline__ void set(unsigned idx, int size)
{
	register int i;
	register unsigned char *p=&cache_index[idx];
	for(i=size;i>0;i--)
		*p++=i;
}

static __inline__ void unset(unsigned idx, int size)
{
	register int i=size;
	register unsigned char *p=&cache_index[idx];
	while(i--)
		*p++=0;
}

void *cache_malloc(size_t real_size)
{
	int hole;
	int size=real_size>>MIN_BITS;
	if (real_size&((1<<MIN_BITS)-1))
		size++;
	if (size<=0)
		size=1;
	if (size>=MAX_SIZE)
		return malloc(real_size);
	hole=get_hole(size);
	if (hole<0)
		return malloc(real_size);
	set(hole,size);
	return (void *)&cache_buffer[((unsigned)hole)<<MIN_BITS];
}

void cache_free(void *ptr)
{
	unsigned idx=(((unsigned)ptr)-((unsigned)&cache_buffer[0]))>>MIN_BITS;
	if (idx>=MAX_INDEX)
		free(ptr);
	else
	if (cache_index[idx])
	{
		if (idx && cache_index[idx-1]>1)
			return;
		unset(idx,cache_index[idx]);
	}
}

void *cache_realloc(void *ptr, size_t real_size)
{
	unsigned idx,back;
	if (!ptr)
		return cache_malloc(real_size);
	if (!real_size)
	{
		cache_free(ptr);
		return NULL;
	}
	idx=(((unsigned)ptr)-((unsigned)&cache_buffer[0]))>>MIN_BITS;
	if (idx>=MAX_INDEX)
		return realloc(ptr,real_size);
	back=cache_index[idx];
	if (back)
	{
		void *newptr;
		unsigned back_cache_actual=cache_actual;
		int size,hole;
		if (idx && cache_index[idx-1]>1)
			return NULL;
		size=real_size>>MIN_BITS;
		if (size<0)
			size=1;
		if (real_size&((1<<MIN_BITS)-1))
			size++;
		if (back==size)
			return ptr;

		unset(idx,back);
		if (back>=size)
		{
			set(idx,size);
			return ptr;
		}
		cache_actual=idx;
		hole=get_hole(size);
		if ((hole<0)||(size>=MAX_SIZE))
			newptr=malloc(real_size);
		else
		if (hole==idx)
		{
			set(idx,size);
			cache_actual=back_cache_actual;
			return ptr;
		}
		else
		{
			set(hole,size);
			newptr=(void *)&cache_buffer[((unsigned)hole)<<MIN_BITS];
		}
		back<<=MIN_BITS;
		if (back<real_size)
			real_size=back;
		memcpy(newptr,ptr,real_size);
		cache_actual=back_cache_actual;
		return newptr;
	}
	return NULL;
}

void *cache_calloc(size_t nmemb, size_t size)
{
	unsigned sz=nmemb*size;
	void *ret=cache_malloc(sz);
	if (ret)
		memset(ret,0,sz);
	return ret;
}

char *cache_strdup(const char *str)
{
	int len;
	char *result;
	if (!str)
		return NULL;
	len = (int)strlen(str);
	result = cache_malloc(len+1);
	if (len>0)
		memcpy(result, str, len);
	result[len] = 0;
	return result;
}

char *cache_strndup(const char *str, size_t size)
{
	int len;
	char *result;
	if (!str)
		return NULL;
	len = (int)strlen(str);
	if (len>size)
		len=size;
	result = cache_malloc(len);
	if (len>0)
		memcpy(result, str, len);
	result[len] = 0;
	return result;
}


#endif
